﻿// ---------------------------------------------------------------------------
// <copyright file="XapProxy.cs" company="Microsoft">
//     Copyright (c) Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>Fiddler Extension that injects a spy dll to the SL app.</summary>
// <owner>Siddhartha Pandey</owner>
// --------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Linq;
using Fiddler;
using Ionic.Utils.Zip;

[assembly: Fiddler.RequiredVersion("2.2.7.5")]

namespace FiddlerXapProxy
{
    /// <summary>
    /// Fiddler extension to intercep Silverlight xap files and inject SilverlightSpy dlls.
    /// </summary>
    public class XapProxy : IAutoTamper
    {
        #region Private Variables

        /// <summary>HTML header used to display messages.</summary>
        private const string SPY_HEADER = "X-SilverlightSpy";

        /// <summary>HTML header used to display error messages.</summary>
        private const string SPY_HEADER_ERROR = SPY_HEADER + "Error";

        /// <summary>List of files to add to xap file.</summary>
        private List<string> _files = new List<string>();

        #endregion
        #region IAutoTamper Members

        /// <summary>
        /// Tamper with request after sending to host.
        /// </summary>
        /// <param name="oSession">Current session.</param>
        [CLSCompliant(false)]
        public void AutoTamperRequestAfter(Session oSession)
        {
        }

        /// <summary>
        /// Tamper with request before sending to host.
        /// </summary>
        /// <param name="oSession">Current session.</param>
        [CLSCompliant(false)]
        public void AutoTamperRequestBefore(Session oSession)
        {
        }

        /// <summary>
        /// Tamper with response after sending to client.
        /// </summary>
        /// <param name="oSession">Current session.</param>
        [CLSCompliant(false)]
        public void AutoTamperResponseAfter(Session oSession)
        {
        }

        /// <summary>
        /// Tamper with response before sending to client.
        /// </summary>
        /// <param name="oSession">Current session.</param>
        [CLSCompliant(false)]
        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public void AutoTamperResponseBefore(Session oSession)
        {
            // Only add dlls if it's a silverlight app and the Spy files exist
            // Don't use header "Content-Type" as some servers dont' have this correctly configured
            if (_files.Count > 0 && oSession.uriContains(".xap") && oSession.responseBodyBytes.Length > 0)
            {
                string tmpFile = string.Format(CultureInfo.InvariantCulture, "{0}\\SLSpy_{1}.xap", Environment.GetEnvironmentVariable("TMP"), Guid.NewGuid().ToString("N"));

                try
                {
                    // Save response from server into temp file
                    int prevSize = oSession.responseBodyBytes.Length;
                    oSession.SaveResponseBody(tmpFile);

                    // Add files and modify appmanifest.xaml
                    InjectSpy(tmpFile);

                    byte[] file = File.ReadAllBytes(tmpFile);

                    if (file != null || file.Length > 0)
                    {
                        oSession.responseBodyBytes = file;
                        oSession.oResponse["Content-Length"] = file.Length.ToString(CultureInfo.InvariantCulture);
                        oSession.oResponse[SPY_HEADER] = string.Format(CultureInfo.InvariantCulture, "Old bytes: {0}, New bytes: {1}", prevSize, file.Length);
                    }
                    else
                    {
                        oSession.oResponse[SPY_HEADER] = "Files not injected.";
                    }
                }
                catch (FileNotFoundException e)
                {
                    if (oSession != null & oSession.oResponse != null)
                    {
                        oSession.oResponse[SPY_HEADER] = e.Message;
                    }
                }
                catch (Exception e)
                {
                    if (oSession != null && oSession.oResponse != null)
                    {
                        oSession.oResponse[SPY_HEADER_ERROR] = e.Message;
                    }
                }
                finally
                {
#if !DEBUG
                    if (File.Exists(tmpFile))
                    {
                        try
                        {
                            //File.Delete(tmpFile);
                        }
                        catch
                        {
                            // Nothing to do, just leave the file in the temp directory. Fix for bug #11474.
                        }
                    }
#endif
                }
            }
        }

        /// <summary>
        /// Invoked before returning error to client.
        /// </summary>
        /// <param name="oSession">Current session.</param>
        [CLSCompliant(false)]
        public void OnBeforeReturningError(Session oSession)
        {
        }

        #endregion
        #region IFiddlerExtension Members

        /// <summary>
        /// Invoked before plugin unload.
        /// </summary>
        public void OnBeforeUnload()
        {
        }

        /// <summary>
        /// Invoked on plugin load.
        /// </summary>
        [SuppressMessage("Microsoft.Globalization", "CA1300:SpecifyMessageBoxOptions")]
        public void OnLoad()
        {
            // Use relative path from the Fiddler installation directory
            string fiddlerPath = "Scripts\\SilverlightSpy\\";

            if (Directory.Exists(fiddlerPath))
            {
                _files.AddRange(Directory.GetFiles(fiddlerPath));
            }
            else
            {
                string title = "Error loading FiddlerXapProxy extension!";
                string msg = "Could not find SilverlightSpy dlls at the expected path: {0}\n\n" +
                                "Fiddler will be unable to capture Silverlight xap files.\n\n" +
                                "Reinstall Silverlight Inspector to fix this problem or remove the FiddlerXapProxy extension.";
            }
        }

        #endregion
        #region Private Methods

        /// <summary>
        /// Adds the SilverlightSpy files to the Xap and modifies it's manifest.
        /// </summary>
        /// <param name="zipFileName">Zip temp file name.</param>
        private void InjectSpy(string zipFileName)
        {
            using (ZipFile zip = new ZipFile(zipFileName))
            {
                string manifestFileName = "AppManifest.xaml";

                // Only add files and modify manifest if AppManifest.xaml exists in the xap
                if (!zip.EntryFilenames.Contains(manifestFileName))
                {
                    throw new FileNotFoundException("Xap file does not contain AppManifest.xaml, skipping injection.");
                }
                else
                {
                    // Update the AppManifest
                    using (MemoryStream ms = new MemoryStream())
                    {
                        // Extract manifest
                        zip.Extract(manifestFileName, ms);
                        ms.Seek(0, SeekOrigin.Begin);

                        // Read manifest
                        StreamReader sr = new StreamReader(ms);
                        string xaml = sr.ReadToEnd();
                        sr.Close();

                        // Update manifest with assembly information
                        Stream output = UpdateManifest(xaml);

                        // Replace existing manifest
                        zip.RemoveEntry(manifestFileName);
                        zip.AddFileStream(manifestFileName, string.Empty, output);
                    }

                    // Add Spy files
                    foreach (string file in _files)
                    {
                        // Don't override any duplicate entries
                        if (!zip.EntryFilenames.Contains(Path.GetFileName(file)))
                        {
                            zip.AddFile(file, string.Empty);
                        }
                    }

                    zip.Save();
                }
            }
        }

        /// <summary>
        /// Edits AppManifest.xaml to add the AsseblyPart nodes for the injected files.
        /// </summary>
        /// <param name="xaml">AppManifest.xaml string.</param>
        /// <returns></returns>
        private static Stream UpdateManifest(string xaml)
        {
            XElement doc = XElement.Parse(xaml);

            XNamespace defaultNamespace = doc.GetDefaultNamespace();
            XElement parts = doc.Element(defaultNamespace + "Deployment.Parts");

            if (parts == null)
            {
                throw new XmlException("Unexpected AppManifest format. Could not find Deployment.Parts node.");
            }

            XNamespace xNamespace = doc.GetNamespaceOfPrefix("x");

            // Add assembly information to Deployment.Parts node
            // Note: Ideally, the namespaces would be read from the assembly itself using reflection when the extension is loaded, 
            // but as these are Silverlight asseblies, .NET throws exceptions when attempting to read these values. If users are 
            // ever allowed to inject multiple files, there should be a text file that can be read that will contan the assembly data.
            parts.Add(CreateAssemblyPart(defaultNamespace, xNamespace, "Microsoft.VisualStudio.TestTools.UITest.Extension.SilverlightUIAutomationHelper", "Microsoft.VisualStudio.TestTools.UITest.Extension.SilverlightUIAutomationHelper.dll"));

            // Include other dependent assemblies referenced by SilverlightHelped.dll and not part of SL runtime.
            foreach(string assemblyName in referencedSlSdkAssemblies)
            {
                if (!xaml.Contains(assemblyName + ".dll"))
                {
                    parts.Add(CreateAssemblyPart(defaultNamespace, xNamespace, assemblyName, assemblyName + ".dll"));
                }
            }

            //// Some xap files already contan this file, ensure we don't add it twice
            //if (!xaml.Contains("System.Xml.Linq.dll"))
            //{
            //    parts.Add(CreateAssemblyPart(defaultNamespace, xNamespace, "System.Xml.Linq", "System.Xml.Linq.dll"));
            //}

#if Needed
            if (!xaml.Contains("System.Runtime.Serialization.dll"))
            {
                parts.Add(CreateAssemblyPart(defaultNamespace, xNamespace, "System.Runtime.Serialization", "System.Runtime.Serialization.dll"));
            }
#endif
            string newXaml = doc.ToString();

            // Write to temp file and then copy over the appmanifest
            MemoryStream ms = new MemoryStream();
            ms.Write(Encoding.ASCII.GetBytes(newXaml), 0, Encoding.ASCII.GetByteCount(newXaml));
            ms.Seek(0, SeekOrigin.Begin);

            return ms;
        }

        /// <summary>
        /// Creates a new AssemblyPart element.
        /// </summary>
        /// <param name="defaultNamespace">Default namespace of XML document.</param>
        /// <param name="xNamespace">Namespace for 'x' namespace.</param>
        /// <param name="codeNamespace">Namespace contained in the assebly (i.e. System.Xml.Linq").</param>
        /// <param name="fileName">Assembly file name.</param>
        /// <returns>Returns a new AsseblyPart elemment with x:Name and Source to the values specified in codeNamespace and fileName.</returns>
        private static XElement CreateAssemblyPart(XNamespace defaultNamespace, XNamespace xNamespace, string codeNamespace, string fileName)
        {
            // Returns an AsseblyPart element: <AssemblyPart x:Name="System.Xml.Linq" Source="System.Xml.Linq.dll" />
            return new XElement(defaultNamespace + "AssemblyPart", new XAttribute(xNamespace + "Name", codeNamespace), new XAttribute("Source", fileName));
        }


        /// <summary>
        /// Initialize list of assemblies required for including in the XAP for Unit testing.
        /// </summary>
        /// <returns></returns>
        private static List<string> InitializeReferencedSlSdkAssemblies()
        {
            List<string> requiredAssemblies = new List<string>();
            requiredAssemblies.Add("System.Windows.Controls");
            requiredAssemblies.Add("System.Windows.Controls.Data");
            requiredAssemblies.Add("System.Windows.Controls.Data.Input");
            requiredAssemblies.Add("System.Windows.Controls.Navigation");
            requiredAssemblies.Add("System.Windows.Controls.Input");
            
            // Add more DLLs required for Unit tests here.
            // Also ensure the corresponding DLLs are specified in Plugins\Silverlight\Spy\CopySLHelper.bat
            // script to be copied in to the Fiddle script directory 

            return requiredAssemblies;
        }

        private static List<string> referencedSlSdkAssemblies = InitializeReferencedSlSdkAssemblies();

        #endregion
    }
}